import std.stdio;
import std.string;
import std.file;
import std.random;
import std.date;
import std.ctype;

import SDL;

import font55;

void error(char[] msg) {
    throw new Error(msg);
}

int rnd(int x) {
    return rand() % x;
}

char[] lfill(char[] s, int n) {
    if (s.length > n) s.length = n;
    else {
        while (s.length != n) {
            s = " " ~ s;
        }
    }
    return s;
}

alias int Side;
const Side P1 = 0;
const Side P2 = 1;

alias int Dir;
const Dir DIR_R = 0;
const Dir DIR_D = 1;
const Dir DIR_L = 2;
const Dir DIR_U = 3;

const BW = 80;
const BW_2 = BW/2;
const BH = 25;

SDL_Surface* scr;
Uint32 black, white, gray;
const SW = 640;
const SH = 480;

void abs2rel_xy(inout int x, inout int y, Side s) {
    if (s == P2) {
        x = BW - x - 1;
        y = BH - y - 1;
    }
}
void abs2rel_op(inout char c, Side s) {
    if (s == P2) {
        switch (c) {
        case '<': c = '>'; break;
        case '>': c = '<'; break;
        case '^': c = 'v'; break;
        case 'v': c = '^'; break;
        default:
        }
    }
}

class Befunge {
    struct Chip {
        char o, c;
        int x, y;
        Uint32 col;
        static Chip opCall(char c0) {
            Chip c;
            c.o = c0;
            c.c = c0;
            return c;
        }
    }

    Chip[][] code;

    this() {
        for (int i = 0; i < BH; i++) {
            Chip[] l;
            for (int j = 0; j < BW; j++) l ~= Chip(' ');
            code ~= l;
        }
    }

    void load(char[] filename, Side side) {
        foreach (j, line; split(cast(char[])std.file.read(filename), "\n")) {
            if (line.length > 40) {
                error(format("%s: more than 40 width", filename));
            }

            if (j >= BH) {
                if (line.length != 0)
                    error(format("%s: more than 25 lines", filename));
                continue;
            }

            foreach (i, c; line) {
                int x = i, y = j;
                char o = c;
                abs2rel_xy(x, y, side);
                abs2rel_op(c, side);

                code[y][x].o = o;
                code[y][x].c = c;
                code[y][x].x = x;
                code[y][x].y = y;
            }
        }
    }

    void initChipColor() {
        for (int j = 0; j < BH; j++) {
            for (int i = 0; i < BW; i++) {
                code[j][i].col = white;
            }
        }
    }

    void draw() {
        if (SDL_MUSTLOCK(scr)) SDL_LockSurface(scr);
        for (int i = 0; i < BW; i++) {
            for (int j = 0; j < BH; j++) {
                char c = code[j][i].c;
                Uint32 col = code[j][i].col;
                if (c < ' ' || c > '~') continue;

                int f = font[c-32];
                if (c == ' ') {
                    f = 4096;
                    if (col == white) col = gray;
                }
                for (int l = 0; l < 5; l++) {
                    for (int k = 0; k < 5; k++) {
                        if (f & 1) {
                            int x = 40 + 1 + k + i * 7;
                            int y = 40 + 1 + l + j * 7;
                            Uint32* p = cast(Uint32*)(cast(Uint8*)scr.pixels + y * scr.pitch + x * scr.format.BytesPerPixel);
                            *p = col;
                        }
                        f >>= 1;
                    }
                }
            }
        }
        if (SDL_MUSTLOCK(scr)) SDL_UnlockSurface(scr);
    }

    // for testing
    void dump_code() {
        foreach (l; code) {
            foreach (c; l) {
                putchar(c.c);
            }
        }
    }
}

class Context {
    enum Mode {
        NORMAL, STRING, STEP, DEAD, WIN
    }

    Befunge bef;
    int x, y;
    Dir dir;
    Side side;
    int[] stack;
    Mode mode;
    char[] output;
    char[] input;
    int turn;
    char[] msg;

    this(Befunge b, Side s, char[] fn) {
        bef = b;
        side = s;

        if (s == P2) {
            dir = DIR_L;
            x = BW-1;
            y = BH-1;
        }

        for (int i = 0; i < 20; i++) {
            output ~= ' ';
            msg ~= ' ';
        }

        msg = lfill(std.string.toupper(fn), 20);
    }

    void die(char[] m) {
        mode = Mode.DEAD;
        msg = lfill(m, 20);
    }

    void pushStdin(char c) {
        if (input.length < 20) {
            input ~= c;
        }
//        if (input.length > 20) input = input[1..$];
    }

    void push(int v) {
        stack ~= v;
        if (stack.length > 20) {
            die("STACK OVER FLOW");
        }
    }

    int pop() {
        if (stack.length == 0) {
            die("STACK UNDER FLOW");
            return -1;
        }
        int r = stack[$-1];
        stack.length = stack.length - 1;
        return r;
    }

    void move() {
        if (mode == Mode.WIN || mode == Mode.DEAD) {
            return;
        }

        turn++;

/*
        if (turn % 10 == 0) {
            msg = msg[1..$] ~ msg[0];
        }
*/

        bef.code[y][x].col = white;

        char c = bef.code[y][x].c;

        switch (mode) {
        case Mode.DEAD:
        case Mode.WIN:
            break;
        case Mode.STEP:
            mode = Mode.NORMAL;
            break;
        case Mode.STRING: {
            if (c == '"') mode = Mode.NORMAL;
            else push(c);
            break;
        }
        case Mode.NORMAL:
            switch (c) {
            case '<': dir = DIR_L; break;
            case '>': dir = DIR_R; break;
            case '^': dir = DIR_U; break;
            case 'v': dir = DIR_D; break;
            case '_':
                if (x < BW_2)
                    dir = (pop() == 0) ? DIR_R : DIR_L;
                else
                    dir = (pop() != 0) ? DIR_R : DIR_L;
                break;
            case '|':
                if (x < BW_2)
                    dir = (pop() == 0) ? DIR_D : DIR_U;
                else
                    dir = (pop() != 0) ? DIR_D : DIR_U;
                break;
            case '?':
                dir = rnd(4);
                break;

            case '#':
                mode = Mode.STEP;
                break;
            case '@':
                die("EXITED");
                break;
            case '"':
                mode = Mode.STRING;
                break;

            case '&':
                break;
            case '~':
                if (input.length == 0) push(0);
                else {
                    push(input[0]);
                    input = input[1..$];
                }
                break;
            case '.':
//                printf("%d", pop());
                break;
            case ',': {
//                putchar(pop());
                output = output[1..$] ~ cast(char)pop();
                break;
            }

            case '+': {
                int y = pop();
                push(pop()+y);
                break;
            }
            case '-': {
                int y = pop();
                push(pop()-y);
                break;
            }
            case '*': {
                int y = pop();
                push(pop()*y);
                break;
            }
            case '/': {
                int y = pop();
                push(pop()/y);
                break;
            }
            case '%': {
                int y = pop();
                push(pop()%y);
                break;
            }
            case '`': {
                int y = pop();
                push(y < pop());
                break;
            }
            case '!': {
                push(!pop());
                break;
            }

            case ':': {
                int y = pop();
                push(y);
                push(y);
                break;
            }
            case '\\': {
                int y = pop();
                int x = pop();
                push(y);
                push(x);
                break;
            }
            case '$':
                pop();
                break;

            case 'g': {
                int y = pop();
                int x = pop();
                abs2rel_xy(x, y, side);
                push(bef.code[y][x].o);
                break;
            }
            case 'p': {
                int y = pop();
                int x = pop();
                abs2rel_xy(x, y, side);
                char o = pop();
                bef.code[y][x].o = o;
                abs2rel_op(o, side);
                bef.code[y][x].c = o;
                break;
            }
            default: {
                if (c >= '0' && c <= '9') push(c-'0');
            }
            }
        }

        switch (dir) {
        case DIR_R: x++; break;
        case DIR_D: y++; break;
        case DIR_L: x--; break;
        case DIR_U: y--; break;
        }
        if (x >= BW || x < 0 || y >= BH || y < 0) {
            die("OUT OF BOUNDS");
        }
        else bef.code[y][x].col = SDL_MapRGB(scr.format, 255, 0, 0);
    }

    void draw() {
        if (SDL_MUSTLOCK(scr)) SDL_LockSurface(scr);

        int x = 320 * side;
        int y = 245;
        int sp1 = 19;
        int sp2 = 28;

        render(scr, "CORE WARS 2D", 320-6*12, 18, white);

        render(scr, "MESSAGE:", x+30, y, white);
        y += sp1;
        render(scr, msg, x+40, y, white);
        y += sp2;

        render(scr, "STACK:", x+30, y, white);
        y += sp1;
        char[] ststr;
        foreach (v; stack) ststr ~= v;
        render(scr, ststr, x+40, y, white);
        y += sp2;

        render(scr, "STDOUT:", x+30, y, white);
        y += sp1;
        render(scr, output, x+40, y, white);
        y += sp2;

        render(scr, "STDIN:", x+30, y, white);
        y += sp1;
        render(scr, lfill(input, 20), x+40, y, white);
        y += sp2;

        render(scr, "STATE:", x+30, y, white);
        y += sp1;
        render(scr, format(" (%2d, %2d)", this.x, this.y), x+40, y, white);
        y += sp2;

        if (SDL_MUSTLOCK(scr)) SDL_UnlockSurface(scr);
    }
}

void initSDL() {
    SDL_Init(SDL_INIT_VIDEO);
    SDL_WM_SetCaption("CORE WARS 2D", null);
    scr = SDL_SetVideoMode(SW, SH, 32, SDL_SWSURFACE);
    black = SDL_MapRGB(scr.format, 0, 0, 0);
    white = SDL_MapRGB(scr.format, 255, 255, 255);
    gray = SDL_MapRGB(scr.format, 128, 128, 128);
}

int main(char[][] args) {
    if (args.length != 3) {
        writefln("usage: %s [1p.bef] [2p.bef]", args[0]);
        return 1;
    }

    rand_seed(getUTCtime(), 0);

    Befunge bef = new Befunge;
    bef.load(args[1], P1);
    bef.load(args[2], P2);
//    bef.dump_code;

    initSDL();

    bef.initChipColor();

    Context[2] players = [ new Context(bef, P1, args[1]),
                           new Context(bef, P2, args[2]) ];

    bool done = false;
    while (!done) {
        SDL_Event ev;
        while (SDL_PollEvent(&ev)) {
            if (ev.type == SDL_QUIT) {
                done = true;
            }
            else if (ev.type == SDL_KEYDOWN) {
                SDLKey k = ev.key.keysym.sym;
                if (k == SDLK_ESCAPE) {
                    done = true;
                }
                if (k >= ' ' &&  k <= '~') {
                    char v = k;
                    if ((ev.key.keysym.mod & 0x3) != 0) {
                        v = std.ctype.toupper(v);
                    }
                    players[0].pushStdin(v);
                }
            }
        }

        foreach (p; players) {
            p.move();
        }

        SDL_FillRect(scr, null, black);
        foreach (p; players) {
            p.draw();
        }
        bef.draw();

        SDL_Flip(scr);

        SDL_Delay(16);
    }

    SDL_Quit();

    return 0;
}


